import React from 'react';
import { noopLogger } from 'typescript-log';
import { getRegistrationCreators } from '.';
import { CompositionRenderer } from './renderers/composition-renderer';
import { CompositionsRenderer } from './renderers/compositions-renderer';
import { composeMiddleware } from './middlewares';
import { ComponentsRenderer } from './renderers/components-renderer';
export function LayoutRegistration() {
    return {
        registerComponents: function (registerCallback) {
            // the internal collection of registered components
            var registeredComponents = {};
            var componentMiddlewares = [];
            var builder = {
                registerComponent: function (registration) {
                    if (registeredComponents[registration.type]) {
                        throw new Error(registration.type + " has already been registered");
                    }
                    registeredComponents[registration.type] = registration;
                    return builder;
                },
                registerMiddleware: function (componentMiddleware) {
                    componentMiddlewares.push(componentMiddleware);
                    return builder;
                },
            };
            registerCallback({
                registerComponent: function (registration) {
                    if (registeredComponents[registration.type]) {
                        throw new Error(registration.type + " has already been registered");
                    }
                    registeredComponents[registration.type] = registration;
                    return builder;
                },
            });
            var log = noopLogger();
            // the following exposes registerCompositions
            return {
                registerCompositions: function (registerCallback) {
                    var registeredCompositions = {};
                    var compositionMiddlewares = [];
                    var componentRegistrations = {
                        isRegistered: function (type) {
                            return registeredComponents[type] !== undefined;
                        },
                        get: function (type) {
                            var foundComponent = registeredComponents[type];
                            if (!foundComponent && process.env.NODE_ENV !== 'production') {
                                // continue rendering in production only. otherwise throw, this is so the site does not crash
                                // empty area will be rendered instead
                                log.warn({ type: type }, "Component has not been registered");
                            }
                            return foundComponent;
                        },
                    };
                    var compositionRegistrations = {
                        get: function (type) {
                            var foundComposition = registeredCompositions[type];
                            if (!foundComposition && process.env.NODE_ENV !== 'production') {
                                // Warn a component is missing if not in production
                                log.warn({ type: type }, "Composition has not been registered");
                            }
                            return foundComposition;
                        },
                    };
                    var createRegisterableComponent = getRegistrationCreators().createRegisterableComponent;
                    var middlewares = {
                        component: function () { return null; },
                        composition: function () { return null; },
                    };
                    var layout = {
                        _componentType: undefined,
                        _compositionType: undefined,
                        componentRegistrations: componentRegistrations,
                        compositionRegistrations: compositionRegistrations,
                        component: function (component) {
                            return component;
                        },
                        composition: function (composition) {
                            return composition;
                        },
                        nestedComposition: function (composition) {
                            return {
                                type: 'nested-composition',
                                props: { composition: composition },
                            };
                        },
                        createRenderers: function (_a) {
                            var services = _a.services, _b = _a.log, userLog = _b === void 0 ? noopLogger() : _b;
                            log = userLog;
                            return {
                                renderComponents: function () {
                                    var components = [];
                                    for (var _i = 0; _i < arguments.length; _i++) {
                                        components[_i] = arguments[_i];
                                    }
                                    return (React.createElement(ComponentsRenderer, { layoutApi: layout, componentMiddleware: middlewares.component, componentRegistrations: componentRegistrations, services: services, components: components, componentRenderPath: "" }));
                                },
                                renderCompositions: function () {
                                    var compositions = [];
                                    for (var _i = 0; _i < arguments.length; _i++) {
                                        compositions[_i] = arguments[_i];
                                    }
                                    return (React.createElement(CompositionsRenderer, { log: log, layoutApi: layout, componentMiddleware: middlewares.component, compositionMiddleware: middlewares.composition, componentRegistrations: componentRegistrations, compositionRegistrations: compositionRegistrations, services: services, compositions: compositions }));
                                },
                            };
                        },
                    };
                    builder.registerComponent(createRegisterableComponent('nested-composition', function (props, services) {
                        if (props.composition.type === 'nested-composition') {
                            throw new Error("nested-composition is registered as a component, it is not allowed within a nested composition as it's not a composition, please check the route");
                        }
                        return (React.createElement(CompositionRenderer, { composition: props.composition, componentRenderPath: props.componentRenderPath + "nested:" + props.composition.type, services: services, layoutApi: layout, componentRegistrations: componentRegistrations, compositionRegistrations: compositionRegistrations, componentMiddleware: middlewares.component, compositionMiddleware: middlewares.composition, log: log }));
                    }));
                    var compositionBuilder = {
                        registerComposition: function (registration) {
                            if (registeredCompositions[registration.type]) {
                                throw new Error(registration.type + " has already been registered");
                            }
                            registeredCompositions[registration.type] = registration;
                            return compositionBuilder;
                        },
                        registerMiddleware: function (compositionMiddleware) {
                            compositionMiddlewares.push(compositionMiddleware);
                            return compositionBuilder;
                        },
                    };
                    registerCallback(compositionBuilder);
                    middlewares.component = composeMiddleware(componentMiddlewares);
                    middlewares.composition = composeMiddleware(compositionMiddlewares);
                    return layout;
                },
            };
        },
    };
}
//# sourceMappingURL=LayoutRegistration.js.map